require 'securerandom'
require 'json'
require 'uu/os/qos/qos_handler'
require 'uu/os/rest/remote_client'
require 'uu/os/uesuri'
require 'uu/os/env/lock/lock_create'
require 'uu/os/env/lock/lock_creation_timeout_exception'
require 'uu/os/env/process'

module UU
  module OS
    module Env

      # Lock component. Lock must be used inside running {UU::OS::Env::Process process}.
      #
      # @example Usage:
      #          UU::OS::Env::Process.run('ues:SYS:SOME_UC', :main_entity_uri => 'ues:UNI:SOME_ARTIFACT') do
      #            ...
      #            lock = UU::OS::Env::Lock.create(resourceUri, createDto)
      #            begin
      #              ...
      #            ensure
      #              lock.unlock
      #            end
      #            ...
      #          end
      class Lock
        protected :initialize

        # Lock service path.
        # @private
        @@SVC_PATH = '/ues/platform/concurrent/UESLockManager'

        # Name of thread parameter with stored owner ID.
        # @private
        @@LOCK_OWNER_ID = :'uu.os.env.lock-owner-id'

        # Default timeout for which can proces wait for lock creation or resource unlock.
        # @private
        @@DEFAULT_WAIT_TIMEOUT = 60 # 1 minute

        # Maximum timeout for which can proces wait for lock creation or resource unlock.
        # @private
        @@MAXIMUM_WAIT_TIMEOUT = 1800 # 30 minutes
        
        # Value returned by REST API in case of null result
        # @private
        @@NULL_URI_VALUE = '{"nil":"true"}'
        
        # Value returned by REST API in case of "true" result
        # @private
        @@TRUE_VALUE = '"true"'

        # Creates new lock.
        #
        # @param [String, UU::OS::UESURI] resource_uri URI of resource to be locked
        # @param [LockCreate] params DTO with additional lock parameters
        #
        # @return [Lock] Instance of lock or nil if lock was not created (e.g. resource is already locked)
        def self.create(resource_uri, params = nil)
          process_uri = UU::OS::Env::Process.active_process_uri
          if !process_uri
            raise 'Cannot create lock because there is no active process for current context. Either make sure code is executed inside managed environment or use appropriate API to initialize new process.'
          end
          dto = Lock::LockCreate.new(params)
          dto.send(:attributes)[:ownerId] = owner_id
          dto.send(:attributes)[:processUri] = process_uri
          dto.wait_timeout = @@DEFAULT_WAIT_TIMEOUT if !dto.wait_timeout
          dto.wait_timeout = @@MAXIMUM_WAIT_TIMEOUT if dto.wait_timeout > @@MAXIMUM_WAIT_TIMEOUT
          payload = dto.to_json
          svc = UU::OS::REST::RemoteClient.new(self, @@SVC_PATH)

          remaining_time = dto.wait_timeout
          wait = 0.5
          lock_uri = nil
          while true
            lock_uri = svc.post(:create, resource_uri, payload)
            break if ((lock_uri) && (lock_uri != @@NULL_URI_VALUE)) || (remaining_time <= 0)
            wait *= 2
            if remaining_time > wait
              remaining_time -= wait
            else
              wait = remaining_time
              remaining_time = 0
            end
            sleep(wait.to_i)
          end
          if (lock_uri) && (lock_uri != @@NULL_URI_VALUE)
            lock_attr = svc.get(:getAttributes, UU::OS::UESURI.new(lock_uri))
            if lock_attr
              return self.new(lock_attr)
            else
              raise "The created lock with URI #{lock_uri} was tampered by concurrent process. Resource #{resource_uri} cannot be reliably locked."
            end
          elsif dto.wait_timeout > 0
            raise UU::OS::Env::LockCreationTimeoutException.new("The lock for resource #{resource_uri} was not created in given time of #{dto.wait_timeout} seconds.")
          else
            return nil
          end
        end

        # Returns (generates new if necessary) owner ID related to active process.
        #
        # @return [String] Lock owner ID
        def self.owner_id
          if !Thread.current[@@LOCK_OWNER_ID]
            Thread.current[@@LOCK_OWNER_ID] = SecureRandom.uuid
          end
          Thread.current[@@LOCK_OWNER_ID]
        end
        private_class_method :owner_id

        # Creates new instance of lock.
        #
        # @param [LockCreate] init_data Lock initialization parameters
        #
        # @private
        def initialize(init_data)
          data = JSON.parse(init_data, :symbolize_names => true)
          @lock_uri = UU::OS::UESURI.new(data[:uri])
        end

        # Checks if lock is valid (e.g. is not expired or was not removed by other process).
        #
        # @return True if lock is valid, else false.
        def valid?
          return false if !@lock_uri
          svc = UU::OS::REST::RemoteClient.new(self.class, @@SVC_PATH)
          svc.add_parameter(:ownerId, self.class.send(:owner_id))
          valid = svc.get(:isValid, @lock_uri)
          return valid == @@TRUE_VALUE
        end

        # Reallocates lock and thus allows prolonging (or shortening)
        # its expiration time. In case lock was already removed (e.g.
        # is already expired, or was removed by other process), exception
        # is thrown.
        #
        # @param [Numeric] lock_timeout New value of lock timeout in seconds (may be nil or -1 to maintain lock timeout initially set on lock create)
        def reallocate(lock_timeout = nil)
          return if !@lock_uri
          svc = UU::OS::REST::RemoteClient.new(self.class, @@SVC_PATH)
          svc.add_parameter(:ownerId, self.class.send(:owner_id))
          svc.add_parameter(:lockTimeout, lock_timeout.to_i) if lock_timeout
          svc.post(:reallocate, @lock_uri)
        end

        # Unlocks lock. In case lock is not valid in time of its
        # unlock exception is thrown.
        def unlock
          return if !@lock_uri
          svc = UU::OS::REST::RemoteClient.new(self.class, @@SVC_PATH)
          svc.add_parameter(:ownerId, self.class.send(:owner_id))
          svc.delete(:unlock, @lock_uri)
          @lock_uri = nil
        end

      end

    end
  end
end
