require 'uu/os/rest/dto'
require 'uu/os/uesuri_builder'
require 'uu/os/uesuri'

module UU
  module OS
    
    # UESURI - UES Uniform Resource Identifier (URI with +ues:+ scheme).
    class UESURI < UU::OS::REST::DTO

      private

      # WARNING for building regex:
      # Escaping regex meta-characters in double quotes needs double backslash, e.g. '\$' ~ "\\$"
      # Concatenation does not affect properly escaped meta-characters, e.g. a = '\$'; "#{a}" is ok.

      # Reserved characters: $ # = & ? : @ [ ] { }
      # @private
      RESERVED = '[^\$#=&\?:@\[\]\{\}]'

      # Reserved characters including space
      # @private
      RESERVEDWITHSPACE = '[^\$#=&\?:@\[\]\{\} ]'

      # Reserved characters escaped by backslash
      # @private
      ESCAPED = '([\\\]\$)|([\\\]#)|([\\\]=)|([\\\]&)|([\\\]\?)|([\\\]:)|([\\\]@)|([\\\]\[)|([\\\]\])|([\\\]\{)|([\\\]\})'

      # URI scheme - alphanum, must begin with letter, may contain dash or underscore
      # @private
      SCHEME = '[a-z,A-Z]([a-z,A-Z,0-9,\-,_])*'

      # Entity code: any character (reserved must be escaped), must not contain space, must contain at least one character
      # @private
      CODE = "((#{RESERVEDWITHSPACE})|(#{ESCAPED}))+"

      # Entity ID: alphanum including dash or underscore, must contain at least one character
      # @private
      ID = '\[[a-z,A-Z,0-9,\-,_]+\]'

      # Expression: begins with '${' or '#{', ends with '}', value is alphanum, must begin with letter, may contain dash or underscore
      # @private
      # EXPR = '(\$|#)\{[a-z,A-Z]([a-z,A-Z,0-9,\-,_])*\}'

      # Expression: any character (reserved must be escaped), must not contain space, must contain at least one character
      # @private
      EXPR = "(\\$|#)\\{((#{RESERVEDWITHSPACE})|(#{ESCAPED}))+\\}"

      # URI part: only code, only ID, code with ID, or expression
      # @private
      PART = "(#{CODE}|#{ID}|#{CODE}#{ID}|#{EXPR})"

      # Use case - same as URI part
      # @private
      USECASE = "(#{CODE}|#{ID}|#{CODE}#{ID}|#{EXPR})"

      # Use case action - alphanum including underscore
      ACTION = '([a-z,A-Z,0-9,_])+'

      # URI parameter: begins with '&', uses pattern 'name=value', name and value can use any character (reserved must be escaped)
      # @private
      PARAMETER = "&((#{RESERVED})|(#{ESCAPED}))+=((#{RESERVED})|(#{ESCAPED}))+"

      # URI fragment: begins with '#', can use any character (reserved must be escaped)
      # @private
      FRAGMENT = "#((#{RESERVED})|(#{ESCAPED}))+"

      # Regular expression for URI evaluation
      # @private
      URI_REGEX = /^#{SCHEME}:((#{PART}@#{PART})|(#{PART}))(:#{PART}){1,2}(:)?(\?(#{USECASE})?(:#{ACTION})?(#{PARAMETER})*)?(#{FRAGMENT})?$/

      # Regular expression for expression evaluation
      # @private
      EXPR_REGEX = /^#{EXPR}$/
      
      # Service path
      PATH = 'uu/os/UESURI';

      public

      # Creates new instance of UESURI.
      #
      # @example Create UESURI from a String
      #   uri = UESURI.new('ues:TER:ART')
      #
      # @example Create UESURI from another UESURI (i.e. create a duplicate)
      #   defensive_copy = UESURI.new(another_uri)
      #
      # @param [String, UESURI] data UESURI String or instance of UESURI
      def initialize(data = nil)
        if data.nil?
          return
        elsif (!data.kind_of?String) && (!data.kind_of?UESURI)
          raise ArgumentError.new("Parameter must be a String or an instance of UESURI, but was #{data.class}.")
        end
        remain = data.to_s
        if remain[/^".*"$/] # It is URI in JSON
          remain = remain[1..-2]
        end
        if !(remain =~ URI_REGEX)
          raise ArgumentError.new("Parameter [#{remain}] is not valid UESURI")
        end
        # get schema
        parts = split_one(remain, ':')
        if parts.size != 2
          raise "UESURI (#{data}) does not define schema"
        end
        @schema = parts[0]
        remain = parts[1]
        # get role (authority)
        parts = split_one(remain, '@')
        if parts.size == 2
          @role = Part.new(parts[0])
          remain = parts[1]
        end
        # get fragment
        parts = split_one(remain, '#', true)
        if parts.size == 2
          @fragment = UESURI.send(:url_decode, parts[1])
          remain = parts[0]
        end
        # get usecase part including parameters
        parts = split_one(remain, '?', true)
        if parts.size == 2
          # separate parameters
          tmp = split_one(parts[1], '&')
          if !tmp[0].empty? # use case exists
            # parse optional action
            uca = split_one(tmp[0], ':', true)
            if (uca.size == 2)
              @action = uca[1] if !uca[1].empty?              
            end
            @use_case = Part.new(uca[0]) if !uca[0].empty?
          end
          if tmp[1] # parameters exists
            @parameters = []
            split_all(tmp[1], '&').each do |param|
              @parameters << Parameter.new(param)
            end
          end
          remain = parts[0]
        end
        # get uri parts
        @parts = []
        parts = split_one(remain, ':')
        while parts[0] do
          @parts << Part.new(parts[0])
          parts = split_one(parts[1]?parts[1]:'', ':')
        end
      end

      # Creates a new UESURI instance with attributes from +data+.
      #
      # @example Create UESURI from a Hash
      #   uri = UESURI.new(:schema => 'ues', :territory_code => 'TER', :artifact_code => 'ART')
      #
      # @param [Hash] data UESURI attributes.
      # @option data [String] :role_code Role code.
      # @option data [String] :role_id Role ID.
      # @option data [String] :territory_code Territory code.
      # @option data [String] :territory_id Territory ID.
      # @option data [String] :territory_expr Territory expression. If an expression is set, code and ID are ignored.
      # @option data [String] :artifact_code Artifact code.
      # @option data [String] :artifact_id Artifact ID.
      # @option data [String] :artifact_expr Artifact expression. If an expression is set, code and ID are ignored.
      # @option data [String] :object_code Object code.
      # @option data [String] :object_id Object ID.
      # @option data [String] :object_expr Object expression. If an expression is set, code and ID are ignored.
      # @option data [String] :use_case_code Use case code.
      # @option data [String] :use_case_id Use case ID.
      # @option data [String] :action Use case action.
      # @option data [Hash] :parameters Use case parameters.
      # @option data [String] :fragment Fragment.
      # @return [UESURI] New instance of UESURI.
      def self.create(data)
        unless minimal_uesuri_parts?(data)
          raise ArgumentError.new("The hash does not contain enough data to represent a minimal UESURI (ues:TER:ART): #{data}")
        end

        builder = UESURIBuilder.new

        builder.set_role_code(data[:role_code])
        builder.set_role_id(data[:role_id])

        if data[:territory_expr]
          builder.set_territory_code(data[:territory_expr])
        else
          builder.set_territory_code(data[:territory_code])
          builder.set_territory_id(data[:territory_id])
        end

        if data[:artifact_expr]
          builder.set_artifact_code(data[:artifact_expr])
        else
          builder.set_artifact_code(data[:artifact_code])
          builder.set_artifact_id(data[:artifact_id])
        end

        if data[:object_expr]
          builder.set_object_code(data[:object_expr])
        else
          builder.set_object_code(data[:object_code])
          builder.set_object_id(data[:object_id])
        end

        builder.set_use_case_code(data[:use_case_code])
        builder.set_use_case_id(data[:use_case_id])

        builder.set_action(data[:action])

        if data[:parameters]
          unless data[:parameters].kind_of?(Hash)
            raise ArgumentError.new("Parameters must be represented as a Hash, not #{data[:parameters].class}")
          end
          data[:parameters].each do |k,v|
            builder.add_parameter(k, v)
          end
        end

        builder.set_fragment(data[:fragment])

        builder.to_uesuri
      end

      # Schema.
      # @return [String]
      def schema
        @schema
      end

      # Role (authority) code.
      # @return [String]
      def role_code
        if @role
          @role.code
        end
      end

      # Role (authority) ID.
      # @return [String]
      def role_id
        if @role
          @role.id
        end
      end

      # Territory code.
      # @return [String]
      def territory_code
        if @parts[0]
          @parts[0].code
        end
      end

      # Territory ID.
      # @return [String]
      def territory_id
        if @parts[0]
          @parts[0].id
        end
      end

      # Artifact code.
      # @return [String]
      def artifact_code
        if @parts[1]
          @parts[1].code
        end
      end

      # Artifact ID.
      # @return [String]
      def artifact_id
        if @parts[1]
          @parts[1].id
        end
      end

      # Object code.
      # @return [String]
      def object_code
        if @parts[2]
          @parts[2].code
        end
      end

      # TODO Redefining method object_id causes warning in JRuby.
      # Should we rename the method or suppress this warning?
      warn_level = $VERBOSE
      $VERBOSE = nil

      # Alias default object_id method to be able to use it in inspect even
      # if object_id method is overriden to return Object ID from UESURI.
      #
      # @private
      alias_method :default_object_id, :object_id

      # Object ID.
      # @return [String]
      def object_id
        if @parts[2]
          @parts[2].id
        end
      end

      $VERBOSE = warn_level

      # Use case code.
      # @return [String]
      def use_case_code
        if @use_case
          @use_case.code
        end
      end

      # Use case ID.
      # @return [String]
      def use_case_id
        if @use_case
          @use_case.id
        end
      end

      # Use case action.
      # @return [String]
      def action
        @action
      end

      # Use case parameters.
      # @return [Array<UU::OS::UESURI::Parameter>]
      def parameters
        @parameters
      end

      # Fragment.
      # @return [String]
      def fragment
        @fragment
      end

      # @private
      def to_s
        if @schema.nil?
          return ''
        end
        uri = "#{@schema}:"
        if @role
          uri << "#{@role.to_s}@"
        end
        if @parts
          @parts.each do |part|
            uri << "#{part}:"
          end
        end
        # remove trailing ':'
        uri = uri[0..-2]
        if (@use_case) || (@action) || (@parameters)
          uri << "?#{@use_case}"
          uri << ":#{@action}" if (!@action.nil?) && (!@action.empty?)
          if @parameters
            @parameters.each do |param|
              uri << "&#{param}"
            end
          end
        end
        if @fragment
          uri << "##{UESURI.send(:url_encode, @fragment)}"
        end
        uri
      end

      # @private
      def to_hash
        # UESURI is never serialized as attribute map, it is always send as +String+.
        # Therefore method +to_hash+ returns +String+ instead of real +Hash+.
        uri = to_s
        uri.empty? ? nil : uri
      end

      # Returns normalized UESURI.
      # @param [UU::OS::UESURI] uri_to_normalize UESURI to be normalized. 
      def self.get_normalized_uri(uri_to_normalize)
        svc = UU::OS::REST::RemoteClient.new(UESURI)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getNormalizedUri', uri_to_normalize)
          return UU::OS::UESURI.new(res)
        end
      end
      
      # Compares instance of UESURI with another UESURI in all their parts. Both UESURI are the same as the
      # corresponding parts contains either the same ID or the same code. UESURI parts aren't comparable
      # when one UESURI part contains only code and the other only ID (or vice versa). When incomparable
      # UESURI parts are compared, exception is thrown.
      # @param [String, UESURI] uesuri UESURI String or instance of UESURI
      # @return [TrueClass,FalseClass] True, when both UESURI contains the same UESURI parts at corresponding places, otherwise false.
      def same?(uesuri)
        return false if uesuri.nil? || (uesuri.kind_of?(String) && uesuri.empty?)
        uesuri = UESURI.new(uesuri) if !uesuri.kind_of?UESURI
        part_equal?(role, uesuri.role) && same_entity?(uesuri) && part_equal?(use_case, uesuri.use_case) && (action == uesuri.action) && parameters_equal?(parameters, uesuri.parameters) && (fragment == uesuri.fragment)
      end

      # Compares instance of UESURI with another UESURI according their territory, artifact and object part.
      # Both UESURI are the same as the corresponding parts contains either the same ID or the same code.
      # UESURI parts aren't comparable when one UESURIPart contain only code and the other only ID (or vice
      # versa). When incomparable UESURI parts are compared, exception is thrown.
      # @param [String, UESURI] uesuri UESURI String or instance of UESURI
      # @return [TrueClass,FalseClass] True, when both UESURI contains the same territory, artifact and object, otherwise false.
      def same_entity?(uesuri)
        return false if uesuri.nil? || (uesuri.kind_of?(String) && uesuri.empty?)
        uesuri = UESURI.new(uesuri) if !uesuri.kind_of?UESURI
        (schema == uesuri.schema) && part_equal?(parts[0], uesuri.parts[0]) && part_equal?(parts[1], uesuri.parts[1]) && part_equal?(parts[2], uesuri.parts[2])
      end

      # Compares two UESURI in all their parts. Both UESURI are the same as the corresponding parts contains
      # either the same ID or the same code. UESURI parts aren't comparable when one UESURI part contains only
      # code and the other only ID (or vice versa). When incomparable UESURI parts are compared, exception is thrown.
      # @param [String, UESURI] first UESURI String or instance of first compared UESURI
      # @param [String, UESURI] second UESURI String or instance of second compared UESURI
      # @return [TrueClass,FalseClass] True, when both UESURI contains the same UESURI parts at corresponding places, otherwise false.
      def self.same?(first, second)
        if first.nil? || second.nil?
          return (first == second)
        elsif (first.kind_of?(String) && first.empty?) || (second.kind_of?(String) && second.empty?)
          return (first == second)
        end
        first = UESURI.new(first)
        first.same?second
      end

      # Compares two UESURI according their territory, artifact and object part. Both UESURI are the same as
      # the corresponding parts contains either the same ID or the same code. UESURI parts aren't comparable
      # when one UESURIPart contain only code and the other only ID (or vice versa). When incomparable UESURI
      # parts are compared, exception is thrown.
      # @param [String, UESURI] first UESURI String or instance of first compared UESURI
      # @param [String, UESURI] second UESURI String or instance of second compared UESURI
      # @return [TrueClass,FalseClass] True, when both UESURI contains the same territory, artifact and object, otherwise false.
      def self.same_entity?(first, second)
        if first.nil? || second.nil?
          return (first == second)
        elsif (first.kind_of?(String) && first.empty?) || (second.kind_of?(String) && second.empty?)
          return (first == second)
        end
        first = UESURI.new(first)
        first.same_entity?second
      end

      # Alias default inspect method to be able to use it in order to get
      # the original value if necessary.
      #
      # @private
      alias_method :default_inspect, :inspect

      # @private
      def inspect
        to_s
      end

      # Support for conversions via {UU::OS::Lang::HashConverter}.
      # @private
      def self.__convert(value) # :nodoc:
        case value
        when UU::OS::UESURI
          return value
        when String && URI_REGEX
          return UU::OS::UESURI.new(value)
        else
          super(value)
        end
      end

      protected

      # @private
      def role
        @role
      end

      # @private
      def parts
        @parts
      end

      # @private
      def use_case
        @use_case
      end

      private

      # Splits string by given character (+char+) in two parts.
      #
      # @param [String] data String to be split
      # @param [String] char Character used for split
      # @param last_occurrence Split by first (default) or last occurrence of +char+
      def split_one(data, char, last_occurrence = false) # 21-1 Z RDocu neni jasne, jak by metoda mela fungovat - neni popsan kontrakt; jde sice o privatni metodu, ale z pohledu udrzovatelnosti kodu je to dulezite
        if !last_occurrence # first occurrence
          parts = data.split(char, 2)
        else # last occurrence
          tmp = data.reverse.split(char, 2)
          if tmp.size == 2
            parts = [tmp[1].reverse, tmp[0].reverse]
          else
            parts = [tmp[0].reverse]
          end
        end
        if parts.size < 2
          parts
        else
          if (char == '#') && (parts[1][0] == '{')
            # we do not have fragment but expression
            # TODO Do we need expressions in fragment?
            # If so, we need to implement similar logic as for escaped characters.
            return [data]
          elsif parts[0][/.*\\$/]
            # part ends with backslash - split character was escaped
            tmp = split_one(parts[1], char, last_occurrence)
            if tmp.size < 2
              [parts[0].concat(char).concat(tmp[0])]
            else
              [parts[0].concat(char).concat(tmp[0]), tmp[1]]
            end
          else
            parts
          end
        end
      end

      # Splits string by the given character to unlimited parts.
      #
      # @param [String] data String to be split
      # @param [String] char Character used for split
      def split_all(data, char)
        tmp = split_one(data, char)
        result = [tmp[0]]
        if tmp[1]
          split_all(tmp[1], char).each do |i|
            result << i
          end
        end
        result
      end

      # Checks whether the given hash contains enough data to represent
      # a minimal UESURI.
      #
      # @return +true+ or +false+
      def self.minimal_uesuri_parts?(hash)
        schema = hash[:schema].eql?('ues') || hash[:schema].nil?
        territory = hash[:territory_code] || hash[:territory_id] || hash[:territory_expr]
        artifact = hash[:artifact_code] || hash[:artifact_id] || hash[:artifact_expr]
        (schema && territory && artifact)
      end

      # Encodes given value with URL-encoding.
      #
      # @param [String] value Value to be encoded
      # @return [String] Encoded value
      # @private
      def self.url_encode(value)
        return if value.nil?
        enc = value.to_s.dup
        # Replace backslash first in order not to replace
        # backslashes of already encoded characters.
        enc.gsub!('\\', '\u005C')
        enc.gsub!(' ', '\u0020')
        enc.gsub!(':', '\u003A')
        enc.gsub!('=', '\u003D')
        enc.gsub!('@', '\u0040')
        enc.gsub!('?', '\u003F')
        enc.gsub!('&', '\u0026')
        enc.gsub!('[', '\u005B')
        enc.gsub!(']', '\u005D')
        enc.gsub!('+', '\u002B')
        if !(enc =~ EXPR_REGEX)
          # In case value is not expression, replace
          # also expression delimiters
          enc.gsub!('#', '\u0023')
          enc.gsub!('$', '\u0024')
          enc.gsub!('{', '\u007B')
          enc.gsub!('}', '\u007D')
        end
        enc
      end
      private_class_method :url_encode

      # Decodes given value with URL-encoding.
      #
      # @param [String] value Value to be decoded
      # @return [String] Decoded value
      # @private
      def self.url_decode(value)
        return if value.nil?
        dec = value.to_s.dup
        dec.gsub!('\u0020', ' ')
        dec.gsub!('\u003A', ':')
        dec.gsub!('\u003D', '=')
        dec.gsub!('\u0040', '@')
        dec.gsub!('\u003F', '?')
        dec.gsub!('\u0026', '&')
        dec.gsub!('\u0023', '#')
        dec.gsub!('\u0024', '$')
        dec.gsub!('\u005B', '[')
        dec.gsub!('\u005D', ']')
        dec.gsub!('\u007B', '{')
        dec.gsub!('\u007D', '}')
        dec.gsub!('\u002B', '+')
        dec.gsub!('\u005C', '\\')
        # Decode also backslash escaping
        dec.gsub!('\\ ', ' ')
        dec.gsub!('\\:', ':')
        dec.gsub!('\\=', '=')
        dec.gsub!('\\@', '@')
        dec.gsub!('\\?', '?')
        dec.gsub!('\\&', '&')
        dec.gsub!('\\#', '#')
        dec.gsub!('\\$', '$')
        dec.gsub!('\\[', '[')
        dec.gsub!('\\]', ']')
        dec.gsub!('\\{', '{')
        dec.gsub!('\\}', '}')
        dec.gsub!('\\+', '+')
        dec
      end
      private_class_method :url_decode

      # @private
      def part_equal?(first, second)
        if (first.nil? || second.nil?)
          first == second
        elsif (first.empty? || second.empty?)
          first.empty? == second.empty?
        elsif (!first.id_empty? && !second.id_empty?)
          first.id == second.id
        elsif (!first.code_empty? && !second.code_empty?)
          first.code == second.code
        else
          raise "UESURI part \"#{first}\" cannot be compared with \"#{second}\", because one of them contains only code and the other contains only ID."
        end
      end

      # @private
      def parameters_equal?(first, second)
        if (first.nil? || second.nil?)
          first == second
        elsif (first.size != second.size)
          false
        else
          # TODO UESAPI returns parameters as Map, but uuAPI as list of Parameter objects
          # So we have problem not only with API consistency, but also with simple comparison.
          tmp = {}
          first.each do |param|
            tmp[param.name] = param.value
          end
          second.each do |param|
            return false if !((tmp.has_key?(param.name)) && (tmp[param.name] == param.value))
          end
          true
        end
      end

      # UESURI part.
      # @private
      class Part

        # Part code.
        attr_reader :code

        # Part ID.
        attr_reader :id

        # Creates new instance of Part.
        #
        # @param [String] data UESURI part string
        def initialize(data)
          parts = data.split(']')[0].split('[')
          @code = UESURI.send(:url_decode, parts[0]) if (parts[0]) && (!parts[0].empty?)
          @id = parts[1] if (parts[1]) && (!parts[1].empty?)
        end

        # @private
        def id_empty?
          return @id.nil? || @id.empty?
        end

        # @private
        def code_empty?
          return @code.nil? || @code.empty?
        end

        # @private
        def empty?
          return id_empty? && code_empty?
        end

        # @private
        def to_s
          part = "#{UESURI.send(:url_encode, @code)}"
          if @id
            part << "[#{@id}]"
          end
          part
        end

      end

      public

      # Use case parameter.
      class Parameter

        # Parameter name.
        # @return [String]
        attr_reader :name

        # Parameter value.
        # @return [String]
        attr_reader :value

        # Creates new instance of Parameter
        #
        # @param [String] data Use case parameter string or name-value pair
        #
        # @private
        def initialize(*data)
          if (data.size == 2)
            @name = data[0]
            @value = data[1]
          else
            parts = data[0].split('=')
            @name = UESURI.send(:url_decode, parts[0])
            @value = UESURI.send(:url_decode, parts[1])
          end
        end

        # @private
        def to_s
          "#{UESURI.send(:url_encode, @name)}=#{UESURI.send(:url_encode, @value)}"
        end

      end

      # Nil object for UESURI. Typical usage is in the uuAPI when it is needed to represent empty/nil value.
      NIL_URI = UESURI.new('ues:[-1]:[-1]')

    end

  end
end
